#-------------------------------------------------------#
#     div_c_dp.sc					#
#-------------------------------------------------------#
#   * Function Description :                            #
#     32-bit complex number division (a+bi)/(c+di)   	#
#     producing complex number quotient as a result. 	#
#							#
#     (a+b*i)      (a*c-b*d)     (a*d+b*c)              #
#     -------  =   ---------  +  --------- * i          #
#     (c+d*i)      (c^2+d^2)     (c^2+d^2)              #
#                                                       #
#   * Design Assumption and Constraints :		#
#     a < c ,  b < d                                    #
#   * Input values :					#
#	   (r2+) = c ( low imag. part of divisor)	#
#	   (r2+) = c ( high imag. part of divisor)	#
#          (r2+) = d ( low real part of divisor)  	#
#          (r2+) = d ( high real part of divisor)	#
#	   (r6+) = a ( low imag. part of dividend)	#
#	   (r6+) = a ( high imag. part of dividend)	#
#          (r6+) = b ( low real part of dividend)  	#
#          (r6+) = b ( high real part of dividend)	#
#   * Return values :                                   #
#          (r5+) = Quotient of real part		#
#          (r5+) = Quotient of imag. part		#
#   * Temporary values :                                #
#          RAM0[0~1] = temporary buffer                 #
#          RAM1[0~3] = temporary buffer                 #
#	   RAM0[3] = c^2+d^2          			#
#	   RAM0[6~d] = multiplied outputs          	#
#	   RAM1[6~9] = added and subtracted outputs     #
#   * Altered Registers	:				#
#          Acc, r0, r1, r2, r3, r4, r5, r6, r7, stack,  #
#	   st, x, y       				#
#   * Computation time :                              	#
#          maximum 826 cycles                           #
#                                                       #
#-------------------------------------------------------#
org	0000
dw	0004
dw	0004
dw	0004
dw	0004
org	0004
#--------------------------------------------------------

    # initial part
        ldi	r2, 11			# pointer of c
        ldi     r5, 11                  # pointer of a_high word
        ldi	r6, 20			# pointer of quotient
        ldi	r1, 20			# pointer of remainder
    # main part
    div_c_dp():
    # get a*c
	ldi	r0, 06  	# output pointer of multiplier
	call	always, @mult_r_dp()
    # get a*d
	ld	-, (r5-)
	ld	-, (r5-)        # move pointer to a_high
        call	always, @mult_r_dp()
    # get b*d
	ld	-, (r2-)
	ld	-, (r2-)
    	call	always, @mult_r_dp()
    # c^2+d^2
	ld	-, (r2-)
	ld	-, (r2-)
        mld     (r5-), (r2-), 0		# d^2
        ld 	-, (r2-)
	mpya	(r5-), (r2), 0          # c^2+d^2
	mpya    (r5), (r2)		# dummy for (add a, p)
        ld      002, a             	# c^2+d^2 -> divisor
    # get b*c
    	call	always, @mult_r_dp()
        ldi	r0, 06			# input X pointer of adder
        ldi	r1, 0a			# input Y pointer of adder
	ldi	r4, 06			# output pointer of adder
    ## calculate real part [ a*c(32 bit) - b*d(32bit) ]
	call	always, @sub_r_dp()
    ## calculate imaginary part [ b*c(32 bit) + a*d(32bit) ]
	call	always, @add_r_dp()
        
    # (a*c-b*d)/(c^2+d^2)
	ldi	r0, 02			# pointer of divisor
	ldi	r4, 06			# pointer of dividend
	call	always, @div_r_dp()	# (a*c-b*d)/(c^2+d^2)
        ldi	r0, 02			# pointer of divisor
        call    always, @div_r_dp()     # (b*c+a*d)/(c^2+d^2)
    # end of div_c_dp()
    	ld	-, (r2+)		# increasing pointer of c, d
    	ld	-, (r2+)		# increasing pointer of c, d
    	ld	pc, stack

#-------------------------------------------------------#
#   Included sub routines                               #
#-------------------------------------------------------#
    div_r_dp():
        ldi	(r3|01), 0e		# subc count number
    # get sign of quotient
        ld	ext7, (r4+)		# dividend_low
        ld	a, (r4)			# dividend_high
        mod	always, abs             # to make positive denom
        ld	(r7|01), a
        ld      (r7|00), ext7
        cmp	a, (r0)                 # check division condition
        bra	n=0, @div_impossible
        mld	(r4+), (r0) 		# divisor * dividend_high
        ld	(r7|10), p   	        # sign of quotient
        ld	a, (r0+)	        # to make positive divisor
        mod	always, abs
        ld	(r3|00), a
	ldi	y, 4000         	# for divisor shift
    	ld	x, (r3|00) 		# load divisor
        ld      ext7, (r7|00)           # dividend low 16bit
        mod	always, shr
        mod	always, shl		# clear LSB 1bit
    div_sub_loop:
        ld	a, (r7|01)		# dividend higher 16bit
	cmp	a, p
        bra	n=1, @div_defficient_remainder
        sub	a, p       		# make partial remainder
        mod	always, inc             # generate quotient
        ld	(r7|01), a		# save acc
    div_defficient_remainder:          	# if divisor>remainder
    	mod	always, shl
        ld	(r7|01), a		# save acc
        ld	a, (r3|01)
        subi	1
        ld	(r3|01), a
        bra	n=1, @div_apply_sign
    	bra	always, @div_sub_loop
    div_impossible:
    	ldi	ext7, 1
    div_apply_sign:
    	ld	(r6), ext7           	# save quotient
    	ld	a, (r7|10)		# check quotient sign
    	and	a, -
    	ld	a, (r6)		      	# final low dividend = quotent
    	bra	n=0, @div_r_dp_ret
    	mod	always, neg             # apply sign
    div_r_dp_ret:
        ld	(r6+), a		# save signed quotent
        ld	pc, stack
#--------------------------------------------------------
    # main part
    mult_r_dp():
	ld	a, st
	ori	a, 0200		# resop, st[9]=1 (mac shift on)
	ld	st, a
    # check sign
        mld	(r5-), (r2-)	# Y_H * X_H
        ld	(r7|10), p	# sign of the result
    # get absolute value of Y
        ld	ext7, (r5+)
        ld	a, (r5+)
        mod	always, abs     # if Y<0, find -Y
        ld	(r7|01), a      # save -Y_H
        ld	(r7|00), ext7   # save -Y_L
        ld	a, (r7|00)
        mod	f, resl
        mod     always, ror     # shift right to make unsigned value
        ld	(r7|00), a      # save -Y_L >> 1
    # get absolute value of X
        ld      ext7, (r2+)
        ld      a, (r2+)
        mod     always, abs      # if X<0, find -X
        ld      (r3|01), a       # save -X_H
        ld      (r3|00), ext7    # save -X_L
        ld      a, (r3|00)
        mod	f, resl
        mod     always, ror      # shift right to make unsigned value
        ld	(r3|00), a       # save -X_L >> 1
    mult_r_dp_cal:
	mld     (r7|00), (r3|00) # acc=0, p=(Y_L>>1)*(X_L>>1)
	ld	(r7|11), p       # higher 16bit p result is initial-
                                 # -value of following mac operation
	mld 	(r7|00), (r3|01) # p=(Y_L>>1)*X_H
	ld	ext7, (r7|11)
        mpya    (r7|01), (r3|00) # p=Y_H*(X_L>>1)
    mid_sum:
        mpya    (r7|01), (r3|00) # dummy for (add a, p)
        mod	always, shl
        ld	x, a
        ldi	y, 1             # shift right 15
        andi	00
    last_mult:
    	mod	f, setop
        mpya    (r7|01), (r3|01)  # p=Y_H*X_H
        mpya    (r5+), (r2+)    # dummy for (add a, p)
        mod	always, inc	# error correction
        ld	(r0+), ext7     # save Result_Low
        ld	(r0+), a        # save Result_High
        ld      a, (r7|10)        
    	and	a, -
        bra     n=0, @pos_out
        ld	-, (r0-)	# dummy for decreasing pointer
    	ld	a, (r0-)
        ld      ext7, (r0)
        mod     always, neg     # if sign = negative -> neg. operation
        ld      (r0+), ext7     # save Result_Low
        ld      (r0+), a        # save Result_High
    pos_out:
    	ld	pc, stack


#--------------------------------------------------------
    add_r_dp():
      	mod	f, resop
      	ld	a, (r0+)   		# Lower word of	X
        add     a, (r1+)                # Lower word of Y
        ld	stack, a        	# save sum of lower word
        ld      a, (r0+)                # Higher word of X
      	bra	l=0, @add_r_dp_higher
      	addi	1                       # add carry
    add_r_dp_higher:
        mod     f, setop 
        add	a, (r1+)                # Higher word of Y
        bra	ov=1, @add_r_dp_overf_sat
    add_r_dp_save:
        ld	(r4+), stack       	# save lower word of result
        bra	always, @add_r_dp_ret
    # Overflow or Underflow Saturation !
    add_r_dp_overf_sat:
        bra	n=0, @add_r_dp_res_pos
        ld 	(r4+), -   		# saturation value : 0x7fffffff
        bra	always, @add_r_dp_ret
    add_r_dp_res_pos:
        ldi	(r4+), 0000             # saturation value : 0x80000000
    add_r_dp_ret:
        ld	(r4+), a		# higher word of output
        ld	pc, stack
#----------------------------------------------------------
    sub_r_dp():
      	mod	f, resop
      	ld	a, (r0+)   		# Lower word of	X
        sub     a, (r1+)                # Lower word of Y
        ld	stack, a        	# save sum of lower word
      	ld	a, (r0+)
      	bra	l=1, @sub_r_dp_higher
      	subi	1                       # sub borrow
    sub_r_dp_higher:
        mod	f, setop                # Higher word of X
        sub	a, (r1+)                # Higher word of Y
        bra	ov=1, @sub_r_dp_overf_sat
    sub_r_dp_save:
        ld	(r4+), stack            # save lower word of result
        bra	always, @sub_r_dp_ret
    # Overflow or Underflow Saturation !
    sub_r_dp_overf_sat:
        bra	n=0, @sub_r_dp_res_pos
        ld 	(r4+), -		# saturation value : 0x7fffffff
        bra	always, @sub_r_dp_ret
    sub_r_dp_res_pos:
        ldi	(r4+), 0000             # saturation value : 0x80000000
    sub_r_dp_ret:
        ld	(r4+), a                # save higher word of result
        ld	pc, stack
